--------------------------------------------------------------
-- Real Era Stop
-- Author: Infixo
-- Mar 9, 2017: Version 1
--		Unfortunately, disconnecting techs doesn't work - the game finds them available to research, same as initial ones from Ancient Era
--		Can delete them completely - the game doesn't break
--		BUT must also delete all related objects as game finds them also available to build
--		Other things to consider:
--			- There's one Government available in Medieval - removed, if Medieval is last Era
--			- Must add Future Civic to finish Civic tree (will make things much easier)
-- Mar 13, 2017: Version 1.2
--		Added support for ATOMIC Era
--		Removed connections to Future Society from Lev3 Governments (like in Vanilla game, you don't have to choose any to progress further)
--------------------------------------------------------------

--------------------------------------------------------------
-- PARAMETER
-- Put the code for the last Era. Use ONLY one of the below numbers:
-- '3' - Medieval
-- '4' - Renaissance
-- '5' - Industrial
-- '6' - Modern
-- '7' - Atomic
--------------------------------------------------------------

INSERT INTO GlobalParameters (Name, Value)
VALUES ('RES_MAX_ERA', '5');

--------------------------------------------------------------
-- PREPARATIONS
--------------------------------------------------------------

-- We will delete all Eras AFTER the last one
CREATE VIEW RESEras AS
SELECT EraType
FROM Eras
WHERE ChronologyIndex > (SELECT Value FROM GlobalParameters WHERE Name = 'RES_MAX_ERA');

CREATE VIEW RESTechnologies AS
SELECT TechnologyType
FROM Technologies
WHERE EraType IN (SELECT EraType FROM RESEras);

CREATE VIEW RESCivics AS
SELECT CivicType
FROM Civics
WHERE EraType IN (SELECT EraType FROM RESEras);

--------------------------------------------------------------
-- CIVILIZATIONS
-- What to do with UU/UB that are going to be removed?
-- America's UB - Film Studio (Broadcast Tower)
-- America's UU - P-51 Mustang (Fighter)
-- Brazil's UU - Minas Geraes (Battleship)
-- England's UU - Sea Dog (Privateer)
-- France's UU - Garde Imperiale
-- Germany's UU - U-Boat (Submarine)
-- Japan's UB - Electronics Factory
--------------------------------------------------------------

--------------------------------------------------------------
-- TECH TREE
--------------------------------------------------------------

-- Connect Future Tech
-- We don't know where Future Tech will be, so we'll add to many Eras and remove unnecessary ones
INSERT INTO TechnologyPrereqs (Technology, PrereqTech)
VALUES
	-- Medieval
	('TECH_FUTURE_TECH', 'TECH_EDUCATION'),
	('TECH_FUTURE_TECH', 'TECH_STIRRUPS'),
	('TECH_FUTURE_TECH', 'TECH_MILITARY_ENGINEERING'),
	('TECH_FUTURE_TECH', 'TECH_CASTLES'),
	-- Renaissance
	('TECH_FUTURE_TECH', 'TECH_SQUARE_RIGGING'),
	('TECH_FUTURE_TECH', 'TECH_ASTRONOMY'),
	('TECH_FUTURE_TECH', 'TECH_BANKING'),
	('TECH_FUTURE_TECH', 'TECH_METAL_CASTING'),
	('TECH_FUTURE_TECH', 'TECH_PRINTING'),
	('TECH_FUTURE_TECH', 'TECH_SIEGE_TACTICS'),
	-- Industrial
	('TECH_FUTURE_TECH', 'TECH_STEAM_POWER'),
	('TECH_FUTURE_TECH', 'TECH_SANITATION'),
	('TECH_FUTURE_TECH', 'TECH_ECONOMICS'),
	('TECH_FUTURE_TECH', 'TECH_RIFLING'),
	-- Modern
	('TECH_FUTURE_TECH', 'TECH_ELECTRICITY'),
	('TECH_FUTURE_TECH', 'TECH_RADIO'),
	('TECH_FUTURE_TECH', 'TECH_CHEMISTRY'),
	('TECH_FUTURE_TECH', 'TECH_REPLACEABLE_PARTS'),
	('TECH_FUTURE_TECH', 'TECH_COMBUSTION'),
	-- Atomic
	('TECH_FUTURE_TECH', 'TECH_COMPUTERS'),
	('TECH_FUTURE_TECH', 'TECH_ADVANCED_FLIGHT'),
	('TECH_FUTURE_TECH', 'TECH_ROCKETRY'),
	('TECH_FUTURE_TECH', 'TECH_NUCLEAR_FISSION'),
	('TECH_FUTURE_TECH', 'TECH_SYNTHETIC_MATERIALS');

-- Move Future Tech to correct Era and set it Cost to be 50% higher than the most expensive tech in Era
UPDATE Technologies
SET EraType = (SELECT EraType FROM Eras WHERE ChronologyIndex = (SELECT Value FROM GlobalParameters WHERE Name = 'RES_MAX_ERA')),
	Cost = 1.5 * (SELECT MAX(Cost) FROM Technologies WHERE EraType = (SELECT EraType FROM Eras WHERE ChronologyIndex = (SELECT Value FROM GlobalParameters WHERE Name = 'RES_MAX_ERA')))
WHERE TechnologyType = 'TECH_FUTURE_TECH';

-- Remove unnecessary Future Tech prereqs - ALL EXCEPT last Era
DELETE FROM TechnologyPrereqs
WHERE
	Technology = 'TECH_FUTURE_TECH' AND
	PrereqTech IN (
		SELECT TechnologyType
		FROM Technologies
		WHERE EraType IN (
			SELECT EraType
			FROM Eras
			WHERE ChronologyIndex <> (SELECT Value FROM GlobalParameters WHERE Name = 'RES_MAX_ERA')));

-- Special case for Medieval Tech connecting to Classical Era tech
INSERT INTO TechnologyPrereqs (Technology, PrereqTech)
VALUES ('TECH_FUTURE_TECH', 'TECH_SHIPBUILDING');
DELETE FROM TechnologyPrereqs
WHERE
	Technology = 'TECH_FUTURE_TECH' AND
	PrereqTech = 'TECH_SHIPBUILDING' AND
	NOT EXISTS (SELECT * FROM GlobalParameters WHERE Name = 'RES_MAX_ERA' AND Value = '3');

-- Remove Prereqs from Techs we'll never get into
DELETE FROM TechnologyPrereqs
WHERE Technology IN (SELECT TechnologyType FROM RESTechnologies);

--------------------------------------------------------------
-- CIVIC TREE
--------------------------------------------------------------

-- need a finishing Civic "Future Society"
INSERT INTO Types (Type, Kind) VALUES ('CIVIC_FUTURE_SOCIETY', 'KIND_CIVIC');
INSERT INTO Civics (CivicType, Name, Cost, Repeatable, Description, EraType, BarbarianFree, UITreeRow, AdvisorType)
VALUES ('CIVIC_FUTURE_SOCIETY', 'LOC_CIVIC_FUTURE_SOCIETY_NAME', 3500, 1, 'LOC_CIVIC_FUTURE_SOCIETY_DESCRIPTION', 'ERA_INFORMATION', 0, 0, 'ADVISOR_GENERIC');

-- Connect Future Civic
-- We don't know where Future Civic will be, so we'll add to many Eras and remove unnecessary ones
INSERT INTO CivicPrereqs (Civic, PrereqCivic)
VALUES
	-- Medieval
	('CIVIC_FUTURE_SOCIETY', 'CIVIC_MERCENARIES'),
	('CIVIC_FUTURE_SOCIETY', 'CIVIC_MEDIEVAL_FAIRES'),
	('CIVIC_FUTURE_SOCIETY', 'CIVIC_GUILDS'),
	('CIVIC_FUTURE_SOCIETY', 'CIVIC_DIVINE_RIGHT'),
	-- Renaissance
	('CIVIC_FUTURE_SOCIETY', 'CIVIC_MERCANTILISM'),
	('CIVIC_FUTURE_SOCIETY', 'CIVIC_THE_ENLIGHTENMENT'),
	-- Industrial
	('CIVIC_FUTURE_SOCIETY', 'CIVIC_NATURAL_HISTORY'),
	('CIVIC_FUTURE_SOCIETY', 'CIVIC_URBANIZATION'),
	('CIVIC_FUTURE_SOCIETY', 'CIVIC_SCORCHED_EARTH'),
	-- Modern
	('CIVIC_FUTURE_SOCIETY', 'CIVIC_CONSERVATION'),
	('CIVIC_FUTURE_SOCIETY', 'CIVIC_NUCLEAR_PROGRAM'),
	('CIVIC_FUTURE_SOCIETY', 'CIVIC_CAPITALISM'),
--	('CIVIC_FUTURE_SOCIETY', 'CIVIC_SUFFRAGE'),
--	('CIVIC_FUTURE_SOCIETY', 'CIVIC_TOTALITARIANISM'),
--	('CIVIC_FUTURE_SOCIETY', 'CIVIC_CLASS_STRUGGLE'),
	-- Atomic
	('CIVIC_FUTURE_SOCIETY', 'CIVIC_CULTURAL_HERITAGE'),
	('CIVIC_FUTURE_SOCIETY', 'CIVIC_RAPID_DEPLOYMENT'),
	('CIVIC_FUTURE_SOCIETY', 'CIVIC_SPACE_RACE'),
	('CIVIC_FUTURE_SOCIETY', 'CIVIC_PROFESSIONAL_SPORTS');

-- Move Future Civic to correct Era and set it Cost to be 50% higher than the most expensive civic in Era
UPDATE Civics
SET EraType = (SELECT EraType FROM Eras WHERE ChronologyIndex = (SELECT Value FROM GlobalParameters WHERE Name = 'RES_MAX_ERA')),
	Cost = 1.5 * (SELECT MAX(Cost) FROM Civics WHERE EraType = (SELECT EraType FROM Eras WHERE ChronologyIndex = (SELECT Value FROM GlobalParameters WHERE Name = 'RES_MAX_ERA')))
WHERE CivicType = 'CIVIC_FUTURE_SOCIETY';

-- Remove unnecessary Future Civic prereqs - ALL EXCEPT last Era
DELETE FROM CivicPrereqs
WHERE
	Civic = 'CIVIC_FUTURE_SOCIETY' AND
	PrereqCivic IN (
		SELECT CivicType
		FROM Civics
		WHERE EraType IN (
			SELECT EraType
			FROM Eras
			WHERE ChronologyIndex <> (SELECT Value FROM GlobalParameters WHERE Name = 'RES_MAX_ERA')));
	
-- Remove Prereqs from Civics we'll never get into
DELETE FROM CivicPrereqs
WHERE Civic IN (SELECT CivicType FROM RESCivics);

-- For Medieval, we're gonna remove Monarchy; otherwise there will be 3 govs to close to Classical ones
-- move temp for Humanism - will be removed with all REN entries in a moment
UPDATE Governments SET PrereqCivic = 'CIVIC_HUMANISM' WHERE GovernmentType = 'GOVERNMENT_MONARCHY' AND EXISTS (SELECT * FROM GlobalParameters WHERE Name = 'RES_MAX_ERA' AND Value = '3');

--------------------------------------------------------------
-- REMOVALS
-- If any of these should be available in the game, it should be connected to an earlier tech/civic
--------------------------------------------------------------

DELETE FROM Buildings WHERE PrereqTech IN (SELECT TechnologyType FROM RESTechnologies);
DELETE FROM Buildings WHERE PrereqCivic IN (SELECT CivicType FROM RESCivics);
-- special case for Rockets - have only District as prereq
DELETE FROM Buildings WHERE PrereqDistrict IN (SELECT DistrictType FROM Districts WHERE PrereqTech IN (SELECT TechnologyType FROM RESTechnologies));
DELETE FROM Buildings WHERE PrereqDistrict IN (SELECT DistrictType FROM Districts WHERE PrereqCivic IN (SELECT CivicType FROM RESCivics));

DELETE FROM Districts WHERE PrereqTech IN (SELECT TechnologyType FROM RESTechnologies);
DELETE FROM Districts WHERE PrereqCivic IN (SELECT CivicType FROM RESCivics);

DELETE FROM Improvements WHERE PrereqTech IN (SELECT TechnologyType FROM RESTechnologies);
DELETE FROM Improvements WHERE PrereqCivic IN (SELECT CivicType FROM RESCivics);

DELETE FROM Policies WHERE PrereqCivic IN (SELECT CivicType FROM RESCivics);

DELETE FROM Governments WHERE PrereqCivic IN (SELECT CivicType FROM RESCivics);

DELETE FROM Projects WHERE PrereqTech IN (SELECT TechnologyType FROM RESTechnologies);
DELETE FROM Projects WHERE PrereqCivic IN (SELECT CivicType FROM RESCivics);

DELETE FROM Resources WHERE PrereqTech IN (SELECT TechnologyType FROM RESTechnologies);
DELETE FROM Resources WHERE PrereqCivic IN (SELECT CivicType FROM RESCivics);

DELETE FROM UnitUpgrades WHERE UpgradeUnit IN (SELECT UnitType FROM Units WHERE PrereqTech IN (SELECT TechnologyType FROM RESTechnologies));
DELETE FROM UnitUpgrades WHERE UpgradeUnit IN (SELECT UnitType FROM Units WHERE PrereqCivic IN (SELECT CivicType FROM RESCivics));
DELETE FROM Units WHERE PrereqTech IN (SELECT TechnologyType FROM RESTechnologies);
DELETE FROM Units WHERE PrereqCivic IN (SELECT CivicType FROM RESCivics);

-- Remove ALL techs after the last Era
DELETE FROM Technologies WHERE EraType IN (SELECT EraType FROM RESEras);
-- Remove ALL civics after the last Era
DELETE FROM Civics WHERE EraType IN (SELECT EraType FROM RESEras);

--------------------------------------------------------------
-- CLEAN-UP
--------------------------------------------------------------

DROP VIEW RESCivics;
DROP VIEW RESTechnologies;
DROP VIEW RESEras;
